@@ -48,8 +48,6 @@
 #include <sys/param.h>
 #endif
 
-#include <stdarg.h>
-
 /* Suppress warnings for __attribute__((warn_unused_result)). */
 #define IGNORE_CALL_RESULT(call) do { if (call) {} } while(0)
 
@@ -160,11 +158,11 @@ typedef enum {
 } update_type;
 
 typedef enum {
-    ADD, DEL, BACK, CUT, CUT_EOF, REPLACE,
+    ADD, DEL, BACK, CUT, CUT_TO_EOF, REPLACE,
 #ifndef DISABLE_WRAPPING
     SPLIT_BEGIN, SPLIT_END,
 #endif
-#ifndef DISABLE_COMMENT
+#ifdef ENABLE_COMMENT
     COMMENT, UNCOMMENT, PREFLIGHT,
 #endif
     JOIN, PASTE, INSERT, ENTER, OTHER
@@ -220,8 +218,10 @@ typedef struct syntaxtype {
 	/* The command with which to lint this type of file. */
     char *formatter;
         /* The formatting command (for programming languages mainly). */
+#ifdef ENABLE_COMMENT
     char *comment;
 	/* The line comment prefix (and postfix) for this type of file. */
+#endif
     colortype *color;
 	/* The colors and their regexes used in this syntax. */
     int nmultis;
@@ -469,6 +469,7 @@ enum
 {
     TITLE_BAR = 0,
     LINE_NUMBER,
+    SELECTED_TEXT,
     STATUS_BAR,
     KEY_COMBO,
     FUNCTION_TAG,
@@ -489,7 +490,7 @@ enum
     USE_MOUSE,
     USE_REGEXP,
     TEMP_FILE,
-    CUT_TO_END,
+    CUT_FROM_CURSOR,
     BACKWARDS_SEARCH,
     MULTIBUFFER,
     SMOOTH_SCROLL,
@@ -519,7 +520,8 @@ enum
     JUSTIFY_TRIM,
     SHOW_CURSOR,
     LINE_NUMBERS,
-    NO_PAUSES
+    NO_PAUSES,
+    AT_BLANKS
 };
 
 /* Flags for the menus in which a given function should be present. */
@@ -558,22 +560,35 @@ enum
 #define CONTROL_RIGHT 0x402
 #define CONTROL_UP 0x403
 #define CONTROL_DOWN 0x404
-#define CONTROL_HOME 0x411
-#define CONTROL_END 0x412
-#define SHIFT_CONTROL_LEFT 0x405
-#define SHIFT_CONTROL_RIGHT 0x406
-#define SHIFT_CONTROL_UP 0x407
-#define SHIFT_CONTROL_DOWN 0x408
-#define SHIFT_CONTROL_HOME 0x413
-#define SHIFT_CONTROL_END 0x414
-#define SHIFT_ALT_LEFT 0x409
-#define SHIFT_ALT_RIGHT 0x40a
-#define SHIFT_ALT_UP 0x40b
-#define SHIFT_ALT_DOWN 0x40c
-#define SHIFT_PAGEUP 0x40d
-#define SHIFT_PAGEDOWN 0x40e
-#define SHIFT_HOME 0x40f
-#define SHIFT_END 0x410
+#define CONTROL_HOME 0x405
+#define CONTROL_END 0x406
+#define SHIFT_CONTROL_LEFT 0x411
+#define SHIFT_CONTROL_RIGHT 0x412
+#define SHIFT_CONTROL_UP 0x413
+#define SHIFT_CONTROL_DOWN 0x414
+#define SHIFT_CONTROL_HOME 0x415
+#define SHIFT_CONTROL_END 0x416
+#define ALT_LEFT 0x421
+#define ALT_RIGHT 0x422
+#define ALT_UP 0x423
+#define ALT_DOWN 0x424
+#define SHIFT_ALT_LEFT 0x431
+#define SHIFT_ALT_RIGHT 0x432
+#define SHIFT_ALT_UP 0x433
+#define SHIFT_ALT_DOWN 0x434
+#define SHIFT_HOME 0x455
+#define SHIFT_END 0x456
+#define SHIFT_PAGEUP 0x457
+#define SHIFT_PAGEDOWN 0x458
+
+#ifdef USE_SLANG
+#ifdef ENABLE_UTF8
+#define KEY_BAD 0xFF  /* Clipped error code. */
+#endif
+#define KEY_FLUSH 0x91  /* User-definable control code. */
+#else
+#define KEY_FLUSH KEY_F0  /* Nonexistent function key. */
+#endif
 
 #ifndef NANO_TINY
 /* An imaginary key for when we get a SIGWINCH (window resize). */
@@ -597,6 +612,9 @@ enum
 /* The default width of a tab in spaces. */
 #define WIDTH_OF_TAB 8
 
+/* The default comment character when a syntax does not specify any. */
+#define GENERAL_COMMENT_CHARACTER "#"
+
 /* The maximum number of search/replace history strings saved, not
  * counting the blank lines at their ends. */
 #define MAX_SEARCH_HISTORY 100
